@@ -1,25 +1,35 @@
+//! Mapping source ranges to/from the HIR and MIR.
+
+use std::rc::Rc;
+
 use either::Either;
 use log::trace;
 use rustc_hir::{
-  intravisit::{self, NestedVisitorMap, Visitor as HirVisitor},
-  BodyId, Expr, ExprKind, MatchSource, Stmt,
+  self as hir,
+  intravisit::{self, Visitor as HirVisitor},
+  itemlikevisit::ItemLikeVisitor,
+  BodyId, Expr, ExprKind, ForeignItem, ImplItem, Item, MatchSource, Node, Param, Stmt,
+  TraitItem,
 };
 use rustc_middle::{
-  hir::map::Map,
+  hir::nested_filter::OnlyBodies,
   mir::{
+    self,
     visit::{
       MutatingUseContext, NonMutatingUseContext, NonUseContext, PlaceContext,
       Visitor as MirVisitor,
     },
-    Body, HasLocalDecls, Location, Place, Terminator, TerminatorKind,
+    Body, HasLocalDecls, TerminatorKind,
   },
   ty::TyCtxt,
 };
-use rustc_span::Span;
+use rustc_span::{Span, SpanData};
 
-use crate::mir::utils::{self, SpanExt};
-
-type HirNode<'hir> = Either<&'hir Expr<'hir>, &'hir Stmt<'hir>>;
+use crate::{
+  block_timer,
+  indexed::impls::LocationDomain,
+  mir::utils::{self, SpanDataExt, SpanExt},
+};
 
 // Collect all the spans for children beneath the visited node.
 // For example, when visiting "if true { 1 } else { 2 }" then we
@@ -29,13 +39,7 @@ struct ChildExprSpans<'tcx> {
   tcx: TyCtxt<'tcx>,
 }
 impl HirVisitor<'hir> for ChildExprSpans<'_> {
-  type Map = Map<'hir>;
-
-  fn nested_visit_map(&mut self) -> NestedVisitorMap<Self::Map> {
-    NestedVisitorMap::None
-  }
-
-  fn visit_expr(&mut self, ex: &Expr) {
+  fn visit_expr(&mut self, ex: &hir::Expr) {
     match ex.kind {
       // Don't take the span for the whole block, since we want to leave
       // curly braces to be associated with the outer statement
@@ -55,7 +59,7 @@ impl HirVisitor<'hir> for ChildExprSpans<'_> {
     }
   }
 
-  fn visit_stmt(&mut self, stmt: &Stmt) {
+  fn visit_stmt(&mut self, stmt: &hir::Stmt) {
     if let Some(span) = stmt.span.as_local(self.tcx) {
       self.spans.push(span);
     }
@@ -70,38 +74,38 @@ pub enum EnclosingHirSpans {
 
 #[derive(Clone, Debug)]
 pub struct HirSpannedNode<'hir> {
-  full: Span,
+  full: SpanData,
   outer: Vec<Span>,
-  node: HirNode<'hir>,
+  node: hir::Node<'hir>,
 }
 
 impl HirSpannedNode<'_> {
   fn get_spans(&self, span_type: EnclosingHirSpans) -> Vec<Span> {
     match span_type {
       EnclosingHirSpans::OuterOnly => self.outer.clone(),
-      EnclosingHirSpans::Full => vec![self.full],
+      EnclosingHirSpans::Full => vec![self.full.span()],
     }
   }
 }
 
 struct HirSpanCollector<'a, 'b, 'hir, 'tcx>(&'a mut Spanner<'b, 'hir, 'tcx>);
 
-impl HirVisitor<'hir> for HirSpanCollector<'_, '_, 'hir, '_> {
-  type Map = Map<'hir>;
-
-  fn nested_visit_map(&mut self) -> NestedVisitorMap<Self::Map> {
-    NestedVisitorMap::None
-  }
-
-  fn visit_expr(&mut self, expr: &'hir Expr<'hir>) {
-    intravisit::walk_expr(self, expr);
-
-    let span = match expr.span.as_local(self.0.tcx) {
-      Some(span) if span != self.0.body_span => span,
+macro_rules! try_span {
+  ($self:expr, $span:expr) => {
+    match $span.as_local($self.0.tcx) {
+      Some(span) if !$self.0.invalid_span(span) => span,
       _ => {
         return;
       }
-    };
+    }
+  };
+}
+
+impl HirVisitor<'hir> for HirSpanCollector<'_, '_, 'hir, '_> {
+  fn visit_expr(&mut self, expr: &'hir hir::Expr<'hir>) {
+    intravisit::walk_expr(self, expr);
+
+    let span = try_span!(self, expr.span);
 
     let inner_spans = match expr.kind {
       ExprKind::Loop(_, _, _, header) => {
@@ -135,21 +139,16 @@ impl HirVisitor<'hir> for HirSpanCollector<'_, '_, 'hir, '_> {
     }
 
     self.0.hir_spans.push(HirSpannedNode {
-      full: span,
+      full: span.data(),
       outer: outer_spans,
-      node: Either::Left(expr),
+      node: Node::Expr(expr),
     });
   }
 
   fn visit_stmt(&mut self, stmt: &'hir Stmt<'hir>) {
     intravisit::walk_stmt(self, stmt);
 
-    let span = match stmt.span.as_local(self.0.tcx) {
-      Some(span) if span != self.0.body_span => span,
-      _ => {
-        return;
-      }
-    };
+    let span = try_span!(self, stmt.span);
 
     let mut visitor = ChildExprSpans {
       spans: Vec::new(),
@@ -159,18 +158,36 @@ impl HirVisitor<'hir> for HirSpanCollector<'_, '_, 'hir, '_> {
     let outer_spans = span.subtract(visitor.spans);
 
     self.0.hir_spans.push(HirSpannedNode {
-      full: span,
+      full: span.data(),
       outer: outer_spans,
-      node: Either::Right(stmt),
+      node: Node::Stmt(stmt),
+    });
+  }
+
+  fn visit_param(&mut self, param: &'hir Param<'hir>) {
+    intravisit::walk_param(self, param);
+
+    let span = match param.span.as_local(self.0.tcx) {
+      Some(span) if !self.0.invalid_span(span) => span,
+      _ => {
+        return;
+      }
+    };
+
+    // TODO: more precise outer spans
+    self.0.hir_spans.push(HirSpannedNode {
+      full: span.data(),
+      outer: vec![span],
+      node: Node::Param(param),
     });
   }
 }
 
-#[derive(Clone, Debug)]
+#[derive(Clone, Debug, PartialEq, Eq)]
 pub struct MirSpannedPlace<'tcx> {
-  pub place: Place<'tcx>,
-  pub span: Span,
-  pub location: Location,
+  pub place: mir::Place<'tcx>,
+  pub span: SpanData,
+  pub location: mir::Location,
 }
 
 struct MirSpanCollector<'a, 'b, 'hir, 'tcx>(&'a mut Spanner<'b, 'hir, 'tcx>);
@@ -178,10 +195,11 @@ struct MirSpanCollector<'a, 'b, 'hir, 'tcx>(&'a mut Spanner<'b, 'hir, 'tcx>);
 impl MirVisitor<'tcx> for MirSpanCollector<'_, '_, '_, 'tcx> {
   fn visit_place(
     &mut self,
-    place: &Place<'tcx>,
+    place: &mir::Place<'tcx>,
     context: PlaceContext,
-    location: Location,
+    location: mir::Location,
   ) {
+    trace!("place={place:?} context={context:?} location={location:?}");
     // Three cases, shown by example:
     //   fn foo(x: i32) {
     //     let y = x + 1;
@@ -209,18 +227,67 @@ impl MirVisitor<'tcx> for MirSpanCollector<'_, '_, '_, 'tcx> {
       }
     };
 
-    let span = match span.as_local(self.0.tcx) {
-      Some(span) if !span.source_equal(&self.0.body_span) => span,
+    let span = try_span!(self, span);
+
+    let spanned_place = MirSpannedPlace {
+      place: *place,
+      location,
+      span: span.data(),
+    };
+    trace!("spanned place: {spanned_place:?}");
+
+    self.0.mir_spans.push(spanned_place);
+  }
+
+  // The visit_statement and visit_terminator cases are a backup.
+  // Eg in the static_method case, if you have x = Foo::bar(), then
+  // then a slice on Foo::bar() would correspond to no places. The best we
+  // can do is at least make the slice on x.
+  fn visit_statement(
+    &mut self,
+    statement: &mir::Statement<'tcx>,
+    location: mir::Location,
+  ) {
+    self.super_statement(statement, location);
+
+    if let mir::StatementKind::Assign(box (lhs, _)) = &statement.kind {
+      let span = try_span!(self, statement.source_info.span);
+      let spanned_place = MirSpannedPlace {
+        place: *lhs,
+        location,
+        span: span.data(),
+      };
+      trace!("spanned place (assign): {spanned_place:?}");
+      self.0.mir_spans.push(spanned_place);
+    }
+  }
+
+  fn visit_terminator(
+    &mut self,
+    terminator: &mir::Terminator<'tcx>,
+    location: mir::Location,
+  ) {
+    self.super_terminator(terminator, location);
+
+    let place = match &terminator.kind {
+      mir::TerminatorKind::Call {
+        destination: Some((place, _)),
+        ..
+      } => *place,
+      mir::TerminatorKind::DropAndReplace { place, .. } => *place,
       _ => {
         return;
       }
     };
 
-    self.0.mir_spans.push(MirSpannedPlace {
-      place: *place,
+    let span = try_span!(self, terminator.source_info.span);
+    let spanned_place = MirSpannedPlace {
+      place,
       location,
-      span,
-    });
+      span: span.data(),
+    };
+    trace!("spanned place (terminator): {spanned_place:?}");
+    self.0.mir_spans.push(spanned_place);
   }
 }
 
@@ -228,8 +295,9 @@ pub struct Spanner<'a, 'hir, 'tcx> {
   pub hir_spans: Vec<HirSpannedNode<'hir>>,
   pub mir_spans: Vec<MirSpannedPlace<'tcx>>,
   pub body_span: Span,
+  pub item_span: Span,
   tcx: TyCtxt<'tcx>,
-  body: &'a Body<'tcx>,
+  body: &'a mir::Body<'tcx>,
 }
 
 impl Spanner<'a, 'hir, 'tcx>
@@ -237,14 +305,24 @@ where
   'tcx: 'hir,
 {
   pub fn new(tcx: TyCtxt<'tcx>, body_id: BodyId, body: &'a Body<'tcx>) -> Self {
-    let hir_body = tcx.hir().body(body_id);
+    let hir = tcx.hir();
+    let hir_body = hir.body(body_id);
+    let owner = hir.body_owner(body_id);
+    let item_span = hir.span_with_body(owner);
+
     let mut spanner = Spanner {
       hir_spans: Vec::new(),
       mir_spans: Vec::new(),
       body_span: hir_body.value.span,
+      item_span,
       tcx,
       body,
     };
+    trace!(
+      "Body span: {:?}, item span: {:?}",
+      spanner.body_span,
+      spanner.item_span
+    );
 
     let mut hir_collector = HirSpanCollector(&mut spanner);
     hir_collector.visit_body(hir_body);
@@ -255,9 +333,15 @@ where
     spanner
   }
 
+  pub fn invalid_span(&self, span: Span) -> bool {
+    span.is_dummy()
+      || span.source_equal(self.body_span)
+      || span.source_equal(self.item_span)
+  }
+
   pub fn find_matching<T>(
-    predicate: impl Fn(Span) -> bool,
-    spans: impl Iterator<Item = (Span, T)>,
+    predicate: impl Fn(SpanData) -> bool,
+    spans: impl Iterator<Item = (SpanData, T)>,
   ) -> impl ExactSizeIterator<Item = T> {
     let mut matching = spans
       .filter(|(span, _)| predicate(*span))
@@ -268,87 +352,80 @@ where
 
   pub fn location_to_spans(
     &self,
-    location: Location,
+    location: mir::Location,
+    location_domain: &Rc<LocationDomain>,
     span_type: EnclosingHirSpans,
   ) -> Vec<Span> {
-    // special case for synthetic locations that represent arguments
-    if location.block.as_usize() == self.body.basic_blocks().len() {
-      return vec![];
-    }
+    let (target_span, stmt) = match location_domain.location_to_arg(location) {
+      Some(place) => (self.body.local_decls[place.local].source_info.span, None),
+      None => (
+        self.body.source_info(location).span,
+        Some(self.body.stmt_at(location)),
+      ),
+    };
 
-    let loc_span = match self.body.source_info(location).span.as_local(self.tcx) {
-      Some(span) => span,
-      None => {
+    let target_span = match target_span.as_local(self.tcx) {
+      Some(span) if !self.invalid_span(span) => span,
+      _ => {
         return vec![];
       }
     };
-    if loc_span == self.body_span || loc_span.is_dummy() {
-      return vec![];
-    }
-
-    let is_return = match self.body.stmt_at(location) {
-      Either::Right(Terminator {
-        kind:
-          TerminatorKind::Return | TerminatorKind::Resume | TerminatorKind::Drop { .. },
-        ..
-      }) => true,
-      _ => false,
-    };
-    if is_return {
-      return vec![];
-    }
 
+    let target_span_data = target_span.data();
     let mut enclosing_hir = Self::find_matching(
-      |span| span.contains(loc_span),
+      |span| span.contains(target_span_data),
       self.hir_spans.iter().map(|span| (span.full, span)),
     )
     .collect::<Vec<_>>();
+    trace!("enclosing_hir={enclosing_hir:#?}");
 
     // Get the spans of the immediately enclosing HIR node
     assert!(
       !enclosing_hir.is_empty(),
-      "Location {location:?} (span {loc_span:?}) had no enclosing HIR nodes"
+      "Location {location:?} (span {target_span:?}) had no enclosing HIR nodes"
     );
     let mut hir_spans = enclosing_hir.remove(0).get_spans(span_type);
-    // trace!(
-    //   "Initial hir node:\n{}\nhas spans: {:?}",
-    //   node_to_string(node),
-    //   hir_spans
-    // );
 
     // Include the MIR span
-    hir_spans.push(loc_span);
+    hir_spans.push(target_span);
 
-    // Add the spans of the first enclosing statement
-    if let Some(hir_span) = enclosing_hir.iter().find(|span| span.node.is_right()) {
-      // trace!(
-      //   "Spans for stmt:\n{}\nare {:?}",
-      //   stmt_to_string(hir_span.node.right().unwrap()),
-      // );
-      hir_spans.extend(hir_span.get_spans(span_type));
+    macro_rules! add_first_where {
+      ($f:expr) => {
+        if let Some(node) = enclosing_hir.iter().find($f) {
+          hir_spans.extend(node.get_spans(span_type));
+        }
+      };
     }
 
-    if let Either::Right(terminator) = self.body.stmt_at(location) {
-      match terminator.kind {
-        TerminatorKind::SwitchInt { .. } => {
-          // If the location is a switch, then include the closest enclosing if or loop
-          if let Some(spans) = enclosing_hir
-            .iter()
-            .filter_map(|span| {
-              matches!(
-                span.node.left()?.kind,
-                ExprKind::If(..) | ExprKind::Loop(..)
-              )
-              .then(move || span.get_spans(span_type))
-            })
-            .next()
-          {
-            // trace!("Switch spans: {:?}", spans);
-            hir_spans.extend(spans);
-          }
-        }
-        _ => {}
-      }
+    // Add the spans of the first enclosing statement
+    add_first_where!(|node| matches!(node.node, Node::Stmt(_)));
+
+    // Include `return` keyword if the location is an expression under a return.
+    add_first_where!(|node| {
+      matches!(
+        node.node,
+        Node::Expr(hir::Expr {
+          kind: hir::ExprKind::Ret(_),
+          ..
+        })
+      )
+    });
+
+    if let Some(Either::Right(mir::Terminator {
+      kind: TerminatorKind::SwitchInt { .. },
+      ..
+    })) = stmt
+    {
+      // If the location is a switch, then include the closest enclosing if or loop
+      add_first_where!(|node| {
+        matches!(
+          node.node,
+          Node::Expr(hir::Expr {
+            kind: ExprKind::If(..) | ExprKind::Loop(..),
+            ..
+          })
+        )
+      });
     }
 
     let format_spans = |spans: &[Span]| -> String {
@@ -362,7 +439,7 @@ where
     trace!(
       "Location {location:?} ({})\n  has loc span:\n  {}\n  and HIR spans:\n  {}",
       utils::location_to_string(location, self.body),
-      format_spans(&[loc_span]),
+      format_spans(&[target_span]),
       format_spans(&hir_spans)
     );
 
@@ -384,47 +461,109 @@ where
       .mir_spans
       .iter()
       .map(|mir_span| (mir_span.span, mir_span));
+    let span_data = span.data();
     let mut contained =
-      Self::find_matching(|mir_span| span.contains(mir_span), spans.clone());
-    let mut containing = Self::find_matching(|mir_span| mir_span.contains(span), spans);
+      Self::find_matching(move |mir_span| span_data.contains(mir_span), spans.clone());
+    let mut containing =
+      Self::find_matching(move |mir_span| mir_span.contains(span_data), spans);
 
-    if contained.len() > 0 {
-      let first = contained.next().unwrap();
+    let mut vec = if let Some(first) = contained.next() {
       contained
         .take_while(|other| other.span.size() == first.span.size())
         .chain([first])
         .collect()
-    } else if containing.len() > 0 {
-      let first = containing.next().unwrap();
+    } else if let Some(first) = containing.next() {
       containing
         .take_while(|other| other.span.size() == first.span.size())
         .chain([first])
         .collect()
     } else {
       Vec::new()
-    }
+    };
+    vec.dedup();
+    vec
   }
 }
 
-fn stmt_to_string(st: &Stmt) -> String {
-  rustc_hir_pretty::to_string(rustc_hir_pretty::NO_ANN, |s| s.print_stmt(st))
-}
-
 fn expr_to_string(ex: &Expr) -> String {
   rustc_hir_pretty::to_string(rustc_hir_pretty::NO_ANN, |s| s.print_expr(ex))
 }
 
-#[allow(dead_code)]
-fn node_to_string(node: HirNode) -> String {
-  match node {
-    Either::Left(ex) => expr_to_string(ex),
-    Either::Right(st) => stmt_to_string(st),
+struct BodyFinder<'tcx> {
+  tcx: TyCtxt<'tcx>,
+  bodies: Vec<(Span, BodyId)>,
+}
+
+impl HirVisitor<'tcx> for BodyFinder<'tcx> {
+  type NestedFilter = OnlyBodies;
+
+  fn nested_visit_map(&mut self) -> Self::Map {
+    self.tcx.hir()
   }
+
+  fn visit_nested_body(&mut self, id: BodyId) {
+    let hir = self.nested_visit_map();
+
+    // const/static items are considered to have bodies, so we want to exclude
+    // them from our search for functions
+    if !hir.body_owner_kind(hir.body_owner(id)).is_fn_or_closure() {
+      return;
+    }
+
+    let body = hir.body(id);
+    self.visit_body(body);
+
+    let hir = self.tcx.hir();
+    let span = hir.span_with_body(hir.body_owner(id));
+    trace!(
+      "Searching body for {:?} with span {span:?} (local {:?})",
+      self
+        .tcx
+        .def_path_debug_str(hir.body_owner_def_id(id).to_def_id()),
+      span.as_local(self.tcx)
+    );
+
+    if !span.from_expansion() {
+      self.bodies.push((span, id));
+    }
+  }
+}
+
+impl ItemLikeVisitor<'tcx> for BodyFinder<'tcx> {
+  fn visit_item(&mut self, item: &'tcx Item<'tcx>) {
+    intravisit::walk_item(self, item);
+  }
+
+  fn visit_impl_item(&mut self, impl_item: &'tcx ImplItem<'tcx>) {
+    intravisit::walk_impl_item(self, impl_item);
+  }
+
+  fn visit_trait_item(&mut self, _trait_item: &'tcx TraitItem<'tcx>) {}
+  fn visit_foreign_item(&mut self, _foreign_item: &'tcx ForeignItem<'tcx>) {}
+}
+
+pub fn find_bodies(tcx: TyCtxt<'tcx>) -> Vec<(Span, BodyId)> {
+  block_timer!("find_bodies");
+  let mut finder = BodyFinder {
+    tcx,
+    bodies: Vec::new(),
+  };
+  tcx.hir().visit_all_item_likes(&mut finder);
+  finder.bodies
+}
+
+pub fn find_enclosing_bodies(
+  tcx: TyCtxt<'tcx>,
+  sp: Span,
+) -> impl Iterator<Item = BodyId> {
+  let mut bodies = find_bodies(tcx);
+  bodies.retain(|(other, _)| other.contains(sp));
+  bodies.sort_by_key(|(span, _)| span.size());
+  bodies.into_iter().map(|(_, id)| id)
 }
 
 #[cfg(test)]
 mod test {
-
   use rustc_data_structures::fx::FxHashSet as HashSet;
   use rustc_middle::mir::BasicBlock;
   use test_log::test;
@@ -480,9 +619,10 @@ mod test {
         let outputs = spanner.span_to_places(input_span);
         let snippets = outputs
           .into_iter()
-          .map(|span| source_map.span_to_snippet(span.span).unwrap())
+          .map(|spanned| source_map.span_to_snippet(spanned.span.span()).unwrap())
           .collect::<HashSet<_>>();
 
+        println!("input_span={input_span:?}");
         compare_sets(&desired.iter().collect::<HashSet<_>>(), &snippets);
       }
     });
@@ -510,7 +650,7 @@ mod test {
       let source_map = tcx.sess.source_map();
       for (input_span, desired) in spans.into_iter().zip(expected) {
         let mut enclosing = Spanner::find_matching(
-          |span| span.contains(input_span),
+          |span| span.contains(input_span.data()),
           spanner.hir_spans.iter().map(|span| (span.full, span)),
         )
         .collect::<Vec<_>>();
@@ -574,6 +714,11 @@ mod test {
 }"#;
     let (input, _ranges) = test_utils::parse_ranges(src, [("`(", ")`")]).unwrap();
     test_utils::compile_body(input, move |tcx, body_id, body_with_facts| {
+      let location_domain = LocationDomain::new(
+        &body_with_facts.body,
+        tcx,
+        tcx.hir().body_owner_def_id(body_id).to_def_id(),
+      );
       let source_map = tcx.sess.source_map();
 
       let spanner = Spanner::new(tcx, body_id, &body_with_facts.body);
@@ -619,11 +764,12 @@ mod test {
       ];
 
       for ((i, j), outp) in pairs {
-        let loc = Location {
+        let loc = mir::Location {
           block: BasicBlock::from_usize(*i),
           statement_index: *j,
         };
-        let spans = spanner.location_to_spans(loc, EnclosingHirSpans::OuterOnly);
+        let spans =
+          spanner.location_to_spans(loc, &location_domain, EnclosingHirSpans::OuterOnly);
         let desired = outp.iter().collect::<HashSet<_>>();
         let actual = spans
           .into_iter()
